import { ApolloClient, NormalizedCacheObject, gql } from '@apollo/client';
import { AbstractThrottledAsyncCache } from '@onomy/hub';

type FrameType = 'Minute' | 'QuarterHour' | 'Hour' | 'QuarterDay' | 'Day' | 'Week';

type ParamsObj = {
  baseDenom: string;
  quoteDenom: string;
  frameType: FrameType;
  from?: number;
  to?: number;
  first?: number;
  skipped?: number;
};

type ParamsType = [ParamsObj];

export const BLANK_ONEX_HISTORY_FRAME = {
  startTime: '0',
  endTime: '0',
  startPrice: '0',
  maxPrice: '0',
  minPrice: '0',
  endPrice: '0',
  volume: '0',
};

export type OnexHistoryFrame = typeof BLANK_ONEX_HISTORY_FRAME;

const ONEX_HISTORICAL_DATA_QUERY = gql`
  query historicalFrames(
    $baseDenom: String!
    $quoteDenom: String!
    $frameType: String!
    $skipped: Int!
    $first: Int
    $from: Int
    $to: Int
  ) {
    historicalFrames(
      first: $first
      skip: $skipped
      where: { base: $baseDenom, quote: $quoteDenom, type: $frameType, startTime_lte: $to }
      orderBy: startTime
      orderDirection: desc
    ) {
      startTime
      endTime
      startPrice
      endPrice
      minPrice
      maxPrice
      volume
    }
  }
`;

export class OnexHistoryFrames extends AbstractThrottledAsyncCache<ParamsType, OnexHistoryFrame[]> {
  private apolloClient: ApolloClient<NormalizedCacheObject>;

  constructor(apolloClient: ApolloClient<NormalizedCacheObject>) {
    super();
    this.apolloClient = apolloClient;
  }

  // eslint-disable-next-line class-methods-use-this
  protected cacheKey([{ baseDenom, quoteDenom, frameType, from, to, first, skipped }]: ParamsType) {
    return [baseDenom, quoteDenom, frameType, from, to, first, skipped].join('-');
  }

  protected async implementation({
    baseDenom,
    quoteDenom,
    frameType,
    from,
    to,
    first,
    skipped,
  }: ParamsObj): Promise<OnexHistoryFrame[]> {
    try {
      const resp: {
        data: {
          historicalFrames: OnexHistoryFrame[];
        };
      } = await this.apolloClient.query({
        query: ONEX_HISTORICAL_DATA_QUERY,
        variables: {
          frameType,
          skipped,
          baseDenom,
          quoteDenom,
          from: from ? Math.round(from) : undefined,
          to: to ? Math.round(to) : Math.round(new Date().getTime() / 1000),
          first,
        },
      });

      const {
        data: { historicalFrames },
      } = resp;

      return historicalFrames;
    } catch (e) {
      console.error('OnexExchangeHistory error', e);
      throw e;
    }
  }
}
