import React, { useCallback, useEffect, useMemo, useState } from 'react';
import BigNumber from 'bignumber.js';
import { useAsyncProcess, useAsyncStatuses } from '@onomy/react-utils';
import {
  useAssetBalance,
  useDetailedBalance,
  useFixedAmountInput,
  useAsset,
} from '@onomy/react-hub';
import { OnexClient } from '@onomy/client';
import {
  useOnexDetailedPoolPrice,
  useOnexClient,
  useOnexOrderBook,
  useFixedAmountFromDisplay,
} from '../hooks';
import { usePriceInputs } from './usePriceInputs';

type OrderType = 'limit' | 'stop';
type OrderDirection = 'buy' | 'sell';

export function useCreateOrderForm({
  baseDenom = 'anom',
  quoteDenom = 'usdt',
  orderDirection = 'buy' as OrderDirection,
  orderType: orderTypeProp = 'limit' as OrderType,
}) {
  const onex = useOnexClient();
  const chain = onex?.getBlockchain() ?? null;
  const [direction, setDirection] = useState<OrderDirection>(orderDirection);

  const [baseAsset, baseAssetStatus] = useAsset(baseDenom, chain);
  const [quoteAsset, quoteAssetStatus] = useAsset(quoteDenom, chain);
  const assetStatus = useAsyncStatuses([baseAssetStatus, quoteAssetStatus]);

  const {
    rate,
    price,
    amountBaseFixed,
    amountQuoteFixed,

    pricePerBase,
    pricePerBaseStatus,

    setPrice,
    setAmountBaseFixed,
    setAmountQuoteFixed,

    reset: resetPriceInputs,
  } = usePriceInputs({ baseAsset, quoteAsset });

  const [stopPrice, setStopPrice] = useState('0');
  const stopPriceFixed = useFixedAmountFromDisplay(stopPrice, quoteAsset);
  const [stopPricePerBase, stopPricePerBaseStatus] = useDetailedBalance(quoteAsset, stopPriceFixed);

  const [currentPricePerBase, currentPricePerBaseStatus] = useOnexDetailedPoolPrice({
    baseAsset,
    quoteAsset,
  });
  const [baseBalance, baseBalanceStatus] = useAssetBalance(baseAsset);
  const [quoteBalance, quoteBalanceStatus] = useAssetBalance(quoteAsset);
  const balanceStatus = useAsyncStatuses([baseBalanceStatus, quoteBalanceStatus]);

  const [orderType, setOrderType] = useState<OrderType>(orderTypeProp);
  const [txResult, setTxResult] = useState<null | Awaited<ReturnType<OnexClient['createOrder']>>>(
    null
  );
  const [attemptTx, txStatus, resetTxStatus] = useAsyncProcess();

  const [amountQuote, amountQuoteStatus] = useDetailedBalance(quoteAsset, amountQuoteFixed);
  const [amountBase, amountBaseStatus] = useDetailedBalance(baseAsset, amountBaseFixed);
  const amountStatus = useAsyncStatuses([amountQuoteStatus, amountBaseStatus]);

  const [bookSell, bookSellStatus] = useOnexOrderBook({ baseAsset, quoteAsset, direction: 'sell' });
  const [bookBuy, bookBuyStatus] = useOnexOrderBook({ baseAsset, quoteAsset, direction: 'buy' });
  const bookStatus = useAsyncStatuses([bookSellStatus, bookBuyStatus]);

  const bidAsset = direction === 'buy' ? quoteAsset : baseAsset;
  const askAsset = direction === 'buy' ? baseAsset : quoteAsset;
  const bidBalance = direction === 'buy' ? quoteBalance : baseBalance;
  const askBalance = direction === 'buy' ? baseBalance : quoteBalance;
  const bidAmountFixed = direction === 'buy' ? amountQuoteFixed : amountBaseFixed;
  const askAmountFixed = direction === 'buy' ? amountBaseFixed : amountQuoteFixed;
  const bidAmount = direction === 'buy' ? amountQuote : amountBase;
  const askAmount = direction === 'buy' ? amountBase : amountQuote;
  const setBidAmountFixed = direction === 'buy' ? setAmountQuoteFixed : setAmountBaseFixed;
  const setAskAmountFixed = direction === 'buy' ? setAmountBaseFixed : setAmountQuoteFixed;

  const [bidAmountInput, setBidAmountInput] = useFixedAmountInput({
    value: bidAmountFixed,
    onChangeValue: setBidAmountFixed,
    meta: bidAsset.meta,
  });

  const [askAmountInput, setAskAmountInput] = useFixedAmountInput({
    value: askAmountFixed,
    onChangeValue: setAskAmountFixed,
    meta: askAsset.meta,
  });

  const inputParams: Parameters<OnexClient['createOrder']>[0] = useMemo(() => {
    if (direction === 'buy') {
      return {
        denomAsk: baseDenom,
        denomBid: quoteDenom,
        orderType,
        amount: amountQuoteFixed,
        rate: [rate[1], rate[0]],
      };
    }
    return {
      denomAsk: quoteDenom,
      denomBid: baseDenom,
      orderType,
      amount: amountBaseFixed,
      rate: [rate[0], rate[1]],
    };
  }, [direction, baseDenom, quoteDenom, orderType, amountQuoteFixed, amountBaseFixed, rate]);

  const validationErrors = useMemo(() => {
    const errors: {
      price?: string;
      amountQuote?: string;
      amountBase?: string;
      amountAsk?: string;
      amountBid?: string;
    } = {};

    const bnPrice = new BigNumber(price);
    const bnAmountQuote = new BigNumber(amountQuoteFixed);
    const bnAmountBase = new BigNumber(amountBaseFixed);

    if (bnPrice.isNaN() || bnPrice.lte(0)) {
      errors.price = 'Must be a number > 0';
    }

    if (bnAmountQuote.isNaN() || bnAmountQuote.lte(0)) {
      errors.amountQuote = 'Must be a number > 0';
    }

    if (bnAmountBase.isNaN() || bnAmountBase.lte(0)) {
      errors.amountBase = 'Must be a number > 0';
    }

    if (direction === 'buy' && bnAmountQuote.gt(quoteBalance.amount.toString())) {
      errors.amountQuote = 'Exceeds available balance';
    }

    if (direction === 'sell' && bnAmountBase.gt(baseBalance.amount.toString())) {
      errors.amountBase = 'Exceeds available balance';
    }

    if (direction === 'buy') {
      if (errors.amountBase) errors.amountAsk = errors.amountBase;
      if (errors.amountQuote) errors.amountBid = errors.amountQuote;
    } else {
      if (errors.amountBase) errors.amountBid = errors.amountBase;
      if (errors.amountQuote) errors.amountAsk = errors.amountQuote;
    }

    return Object.keys(errors).length > 0 ? errors : null;
  }, [
    direction,
    price,
    amountQuoteFixed,
    amountBaseFixed,
    quoteBalance.amount,
    baseBalance.amount,
  ]);

  const loadingStatus = useAsyncStatuses([assetStatus, balanceStatus, currentPricePerBaseStatus]);
  const canSubmit = !validationErrors && loadingStatus.finished;

  const submitTx = useCallback(
    (e?: React.FormEvent) => {
      e?.preventDefault();
      if (!canSubmit) return Promise.resolve();
      resetTxStatus();
      setTxResult(null);
      return attemptTx(async () => setTxResult((await onex?.createOrder(inputParams)) ?? null));
    },
    [canSubmit, inputParams, resetTxStatus, attemptTx, onex]
  );

  const reset = useCallback(() => {
    resetPriceInputs();
    setOrderType('limit');
    resetTxStatus();
    setTxResult(null);
  }, [resetPriceInputs, resetTxStatus]);

  useEffect(reset, [baseDenom, quoteDenom, orderDirection, orderType, reset]);

  return {
    txStatus,
    txResult,
    assetStatus,
    balanceStatus,
    loadingStatus,
    amountStatus,
    pricePerBaseStatus,
    stopPricePerBaseStatus,
    currentPricePerBaseStatus,
    bookBuyStatus,
    bookSellStatus,
    bookStatus,

    validationErrors,
    canSubmit,

    baseAsset,
    baseBalance,
    quoteAsset,
    quoteBalance,

    amountQuoteFixed,
    amountBaseFixed,
    baseDenom,
    quoteDenom,
    direction,
    price,
    stopPrice,
    orderType,

    pricePerBase,
    stopPricePerBase,
    currentPricePerBase,

    amountQuote,
    amountBase,

    bookSell,
    bookBuy,

    bidAsset,
    askAsset,
    bidBalance,
    askBalance,
    bidAmount,
    askAmount,

    bidAmountInput,
    askAmountInput,

    setBidAmountInput,
    setAskAmountInput,

    setDirection,
    setPrice,
    setOrderType,
    setAmountQuoteFixed,
    setAmountBaseFixed,

    submitTx,
    reset,
  };
}

export type CreateOrderFormState = ReturnType<typeof useCreateOrderForm>;
