import React, { createContext, useContext } from 'react';
import { useClaimRewardForm, useDelegation } from '@onomy/react-hub';
import { useAsyncStatuses } from '@onomy/react-utils';
import { BlockchainAddress } from '@onomy/frontend-relay-types';
import { PlaceholderErrorDisplay } from '../../components/PlaceholderErrorDisplay';

function useValidatorDetailState({ validatorAddress }: { validatorAddress: BlockchainAddress }) {
  const [delegation, delegationStatus] = useDelegation(validatorAddress);
  const claimRewards = useClaimRewardForm({ validatorAddress });

  const loadingStatus = useAsyncStatuses([delegationStatus, claimRewards.loadingStatus]);

  const validatorDetailRoute = `/wallet/staking/${validatorAddress.chain.name}/${validatorAddress.id}`;

  return {
    delegation,
    claimRewards,

    loadingStatus,
    validatorDetailRoute,
  };
}

export type ValidatorDetailContextType = ReturnType<typeof useValidatorDetailState>;

const ValidatorDetailContext = createContext<ValidatorDetailContextType>(null as any);

export function useValidatorDetail() {
  const state = useContext(ValidatorDetailContext);
  if (!state) throw new Error('ValidatorDetailContext missing');
  return state;
}

function defaultRenderError(error: any) {
  return <PlaceholderErrorDisplay error={error} />;
}

export function ValidatorDetailProvider({
  children,
  renderLoading,
  renderError = defaultRenderError,
  ...props
}: {
  children: JSX.Element | JSX.Element[];
  renderLoading?: () => JSX.Element | JSX.Element[] | null;
  renderError?: typeof defaultRenderError;
} & Parameters<typeof useValidatorDetailState>[0]) {
  const state = useValidatorDetailState(props);

  if (renderLoading && state.loadingStatus.pending) return renderLoading();
  if (state.loadingStatus.error) return renderError(state.loadingStatus.error);
  if (renderLoading && !state.delegation) return renderLoading();
  return (
    <ValidatorDetailContext.Provider value={state}>{children}</ValidatorDetailContext.Provider>
  );
}
